import os
import sys

from django.db import connections
from django.test.runner import ParallelTestSuite, DiscoverRunner

from .models import dynamodb_model_classes


def init_pynamodb_test_prefix(prefix=None):
    for model_class in dynamodb_model_classes:
        model_class._connection = None
        connection = model_class._get_connection()
        if prefix:
            connection.table_name = 'test_{}_{}'.format(prefix, connection.table_name)
        else:
            connection.table_name = 'test_{}'.format(connection.table_name)


def remove_pynamodb_table(model_class):
    if model_class.exists():
        model_class.delete_table(wait=True)


def recreate_pynamodb_table(model_class):
    remove_pynamodb_table(model_class)
    model_class.create_table(wait=True)


_worker_id = 0


def _init_worker(counter):
    global _worker_id

    with counter.get_lock():
        counter.value += 1
        _worker_id = counter.value

    for alias in connections:
        connection = connections[alias]
        settings_dict = connection.creation.get_test_db_clone_settings(str(_worker_id))
        # connection.settings_dict must be updated in place for changes to be
        # reflected in django.db.connections. If the following line assigned
        # connection.settings_dict = settings_dict, new threads would connect
        # to the default database instead of the appropriate clone.
        connection.settings_dict.update(settings_dict)
        connection.close()

    init_pynamodb_test_prefix(_worker_id)


class DynamoDBParallelTestSuite(ParallelTestSuite):

    init_worker = _init_worker


class DynamoDBTestSuiteMixin:

    parallel_test_suite = DynamoDBParallelTestSuite

    def log(self, msg):
        sys.stderr.write(msg + os.linesep)

    def _teardown_pynamodb_database(self, prefix=None):
        init_pynamodb_test_prefix(prefix)
        table_names = []
        for model_class in dynamodb_model_classes:
            table_names.append(model_class._connection.table_name)
            remove_pynamodb_table(model_class)
        self.log('Remove DynamoDB tables ({})...'.format(
            ', '.join("'{}'".format(table_name) for table_name in table_names))
        )

    def teardown_databases(self, old_config, **kwargs):
        super().teardown_databases(old_config, **kwargs)
        if self.parallel > 1:
            for i in range(self.parallel):
                self._teardown_pynamodb_database(str(i + 1))
        else:
            self._teardown_pynamodb_database()

    def _setup_pynamodb_database(self, prefix=None):
        init_pynamodb_test_prefix(prefix)
        table_names = []
        for model_class in dynamodb_model_classes:
            table_names.append(model_class._connection.table_name)
            recreate_pynamodb_table(model_class)
        self.log('Setup DynamoDB tables ({})...'.format(
            ', '.join("'{}'".format(table_name) for table_name in table_names))
        )

    def setup_databases(self, **kwargs):
        databases = super().setup_databases(**kwargs)

        if self.parallel > 1:
            for i in range(self.parallel):
                self._setup_pynamodb_database(str(i + 1))
        else:
            self._setup_pynamodb_database()
        return databases


class DynamoDBTestDiscoverRunner(DynamoDBTestSuiteMixin, DiscoverRunner):
    pass



from pynamodb.attributes import Attribute
from pynamodb.constants import STRING


class StringJoinAttribute(Attribute):

    attr_type = STRING

    def __init__(self, fields, separator='||', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._fields = fields
        self._separator = separator

    def serialize(self, value):
        if isinstance(value, str):
            return value

        serialized_values = []
        for v, f in zip(value, self._fields):
            serialized_values.append(f.serialize(v) or '')
        return self._separator.join(serialized_values)

    def deserialize(self, value):
        serialized_values = value.split(self._separator)
        deserialized_values = []
        for v, f in zip(serialized_values, self._fields):
            deserialized_values.append(f.deserialize(v))
        return deserialized_values


class BooleanUnicodeAttribute(Attribute):

    attr_type = STRING

    def serialize(self, value):
        if value is None:
            return None
        return str(value)

    def deserialize(self, value):
        return value == 'True'



from .models import dynamodb_model_classes


class DynamoDBTestMixin:

    def tearDown(self):
        super().tearDown()
        for model_class in dynamodb_model_classes:
            with model_class.batch_write() as batch:
                for item in model_class.scan():
                    batch.delete(item)



from setuptools import setup, find_packages


setup(
    name='pydjamodb',
    version='0.0.5',
    description="Django interface to PyDjamoDB.",
    keywords='django, DynamoDB, PyDjamoDB',
    author='Lubos Matl',
    author_email='matllubos@gmail.com',
    url='https://github.com/druids/pydjamodb',
    license='MIT',
    package_dir={'pydjamodb': 'pydjamodb'},
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Czech',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
    ],
    install_requires=[
        'django>=2.0',
    ],
    zip_safe=False
)
